#!/usr/bin/python3

import numpy as np
import copy
from qckt.qException import QCktException

####################################################################################################
## Utility functions ###############################################################################
####################################################################################################

def _qbit_realign_full_list(nqbits, qbit_list):
	reord_list = copy.deepcopy(qbit_list)
	for i in reversed(range(nqbits)): # reversed to maintain significance order of the other qbits; poetic correctness :-)
		if i not in reord_list:
			reord_list.append(i)
	return reord_list

def _shuffled_count(nqbits, bitorder):
	shuffled = []
	for i in range(2**nqbits):
		shfval = 0
		for b in range(nqbits):
			dstbit = bitorder[nqbits-b-1]  ## (nqbits-1) - b  becasue bitorder is array like [MSB, ...,LSB]
			shfval += (((i >> b) & 0x1) << dstbit)
		shuffled.append(shfval)
	return shuffled

def _rmat_invrmat(nqbits, qbit_reorder): ## why this name? rmat is matrix that reorders the state vector, invrmat is inverse of rmat
	# this is per counting with the given bit ordering
	shfl_svec_idxs = _shuffled_count(nqbits, qbit_reorder)
	## create the rmat and invrmat
	imat = np.matrix(np.eye(2**nqbits))
	rmat = np.matrix(np.eye(2**nqbits))
	invrmat = np.matrix(np.eye(2**nqbits))
	for i in range(2**nqbits):
		s = shfl_svec_idxs[i]
		rmat[i] = imat[s]
		invrmat[s] = imat[i]
	return (rmat, invrmat)

def _aligned_op(nqbits, oper, qbit_list):
	qbit_reorder = _qbit_realign_full_list(nqbits, qbit_list)
	(rmat,invrmat) = _rmat_invrmat(nqbits, qbit_reorder)
	a_op = invrmat * oper * rmat
	return a_op


def stretched_opmatrix(nqbits,oper,qbit_list):
	c_op = np.kron(oper,np.eye(2**(nqbits-len(qbit_list))))
	a_op = _aligned_op(nqbits,c_op,qbit_list)
	return a_op

def combine_opmatrices_par(op_list):
	res = None
	first = True
	for op in op_list:
		mat = op
		if first:
			res = op
			first = False
		else:
			res = np.kron(res,op)
	return res

def combine_opmatrices_seq(op_list):
	d = (op_list[0]).shape[0]
	res = np.matrix(np.eye(d),dtype=complex)
	for op in op_list:
		r = op.shape[0]
		c = op.shape[1]
		if r != c:
			errmsg = "Operation is not a square matrix."
			raise QCktException(errmsg)
		if r != d:
			errmsg = "Operation matrices not the same size."
			raise QCktException(errmsg)
		res = op*res # remember order of multiplication is opposite of the visual order
	return res

def isunitary(mat):
	floaterr = 10**(-8)
	if type(mat) is not np.matrix:
		return False
	(r,c) = mat.shape
	if r != c:
		return False
	invmat = np.conjugate(np.transpose(mat))
	pmat = np.asarray(mat * invmat)
	for i in range(r):
		for j in range(c):
			if i != j:
				if np.absolute(pmat[i][j]) > floaterr:
					return False
			else:
				if np.absolute(pmat[i][j]-1.0) > floaterr:
					return False
	return True

def opmat_dagger(opMat):
	mat = opMat[:]
	invmat = np.conjugate(np.transpose(mat))
	return invmat

## Utility function to add control bit
def CTL(opMatrix):
	(r,c) = opMatrix.shape
	oparr = np.array(opMatrix)
	coparr = np.eye(r*2,dtype=complex)
	for i in range(r,r*2):
		for j in range(r,r*2):
			coparr[i][j] = oparr[i-r][j-r]
	return np.matrix(coparr,dtype=complex)




import qckt

## Noise
# noise channel is specified as a sequence of pairs of list of kraus operators and probability factor. 
# NoiseChannel has three fields
#   [ (op1,op1_prob_multiplier), (op2,op2_prob_multiplier), ...  ],
#	name,
#   num_target_qubits
#
#	op is in the same format as gates, i.e., ['name', opmatrix]
#		note that mostly the opmatrix is unitary, but cases such as Amplitude Damping is not [[1.0, 0.0],[0.0, sqrt(gamma)]]
#	op_prob_multiplier gives the noise contribution by this op, i.e., = op_prob_multiplier * (op * state * op_dagger)
#
# so, in math terms, rho = op1_prob_multipier * (op1 * rho * op1_dag) + op2_prob_multiplier * (op2 * rho * op2.dag) + ...
#
#   num_target_qubits is the number of qubits this noise channel acts on
# 

class NoiseChannel:
    def __init__(self, name, noise_chan, nqubits):
        self.name = name
        self.noise_chan = noise_chan
        self.nqubits = nqubits

    def __iter__(self):
        self.it = iter(self.noise_chan)
        return self.it
    def __next__(self):
        return next(self.it)

    def __str__(self):
        return f'name={self.name}\nnoise_chan={self.noise_chan}\nnqubits={self.nqubits}'

class NoiseChannelSequence:
    def __init__(self, *args):
        self.noise_chan_sequence = []
        for a in args:
            if a is None:
                pass
            elif type(a) is NoiseChannel:
                self.add(a)
            elif type(a) is NoiseChannelSequence:
                self.extend(a)
            else:
                raise qckt.QCktException('ERROR: Invalid argument, only NoiseChannel and NoiseChannelSequence objects or None expected')
        self.name = self.__name__()

    def extend(self, noise_chan_sequence):
        if noise_chan_sequence is None:
            addition = []
        elif type(noise_chan_sequence) is NoiseChannelSequence:
            addition = noise_chan_sequence.noise_chan_sequence
        else:
            raise qckt.QCktException('ERROR: NoiseChannelSequence object or None expected')
        self.noise_chan_sequence.extend(addition)
        self.name = self.__name__()
        return self

    def add(self, noise_chan):
        if noise_chan is None:
            pass
        elif type(noise_chan) is NoiseChannel:
            self.noise_chan_sequence.append(noise_chan)            
        else:
            raise qckt.QCktException('ERROR: NoiseChannel object expected')
        self.name = self.__name__()
        return self

    def __iter__(self):
        self.it = iter(self.noise_chan_sequence)
        return self.it
    def __next__(self):
        return next(self.it)

    def __name__(self):
        return ','.join([chan.name for chan in self.noise_chan_sequence])

    def __str__(self):
        return self.name

class NoiseChannelApplierSequense:
    def __init__(self, noise_chan=None, qubit_list=None):
        self.name = ''
        self.noise_chan_applier_sequence = []
        if noise_chan is not None:
            if qubit_list is None:
                raise qckt.QCktException('ERROR: qubits list required.')
            self.add(noise_chan, qubit_list)

    def add(self, noise_chan, qbit_list):
        if noise_chan is None:
            addition = []
        elif type(noise_chan) is NoiseChannelSequence:
            addition = [(chan,qbit_list) for chan in noise_chan]
        elif type(noise_chan) is NoiseChannel:
            addition = [(noise_chan, qbit_list)]
        else:
            raise qckt.QCktException('ERROR: NoiseChannelSequence, NoiseChannel object or None expected.')
        self.noise_chan_applier_sequence.extend(addition)
        self.name = self.__name__()
        return self

    def extend(self, noise_chan_applier_seq):
        if noise_chan_applier_seq is None:
            addition = []
        elif type(noise_chan_applier_seq) is NoiseChannelApplierSequense:
            addition = noise_chan_applier_seq.noise_chan_applier_sequence
        else:
            raise qckt.QCktException('ERROR: NoiseChannelApplierSequense object or None expected')
        self.noise_chan_applier_sequence.extend(addition)
        self.name = self.__name__()
        return self

    def new_with_filtered_qubits(self, qubit_list):
        new_noise_chan_applier_seq = NoiseChannelApplierSequense()
        for noise_channel,qblist in self.noise_chan_applier_sequence:
            filtered_qbits = [q for q in qblist if q in qubit_list]
            if len(filtered_qbits) > 0:
                new_noise_chan_applier_seq.add(noise_channel,filtered_qbits)
        return new_noise_chan_applier_seq

    def __name__(self):
        return ','.join([f'({chan.name},{qbits})' for chan,qbits in self.noise_chan_applier_sequence])

    def __iter__(self):
        self.it = iter(self.noise_chan_applier_sequence)
        return self.it
    def __next__(self):
        return next(self.it) 

    def __str__(self):
        return f'{self.name}'


class NoiseProfile:
    def __init__(self, noise_chan_init=None, noise_chan_qubits=None, noise_chan_allgates=None, noise_chan_allsteps=None):
        self.noise_chan_init = noise_chan_init
        self.noise_chan_qubits = noise_chan_qubits
        self.noise_chan_allgates = noise_chan_allgates
        self.noise_chan_allsteps = noise_chan_allsteps
    def get(self, field, defval):
        print(f'deprecated dict.get(key,default) style access, key={field}.')
        retval = defval
        if hasattr(self, field):
            retval = getattr(self, field)
        return retval
    def __getitem__(self, index):
        print(f'deprecated dict[key] style access, key={index}.')
        return self.get(index,None)
    def keys(self):
        print('deprecated dict.keys() access')
        return ['noise_chan_init', 'noise_chan_qubits', 'noise_chan_allgates', 'noise_chan_allsteps']
    

## create a NoiseChannelApplierSequence combining all components of noise to be applied when gate is applied on certain qubit_list
#
def consolidate_gate_noise(noise_profile, gate_noise, qubit_list):
    noise_chan_applier_sequence_overall = NoiseChannelApplierSequense()

    # 1. noise on specific qubits, noise_chan_qubits
    noise_chan_applier_sequence_qubits = NoiseChannelApplierSequense()
    if noise_profile is not None:
        noise_profile_qubits = noise_profile.noise_chan_qubits
        if noise_profile_qubits is not None:
            noise_chan_applier_sequence_qubits = noise_profile_qubits.new_with_filtered_qubits(qubit_list)
    noise_chan_applier_sequence_overall.extend(noise_chan_applier_sequence_qubits)
    # 2. noise on all gates, noise_chan_allgates
    noise_chan_applier_sequence_all_gates = NoiseChannelApplierSequense()
    if noise_profile is not None:
        noise_profile_allgates = noise_profile.noise_chan_allgates
        noise_chan_applier_sequence_all_gates = NoiseChannelApplierSequense(noise_chan=noise_profile_allgates, qubit_list=qubit_list)
    noise_chan_applier_sequence_overall.extend(noise_chan_applier_sequence_all_gates)
    # 3. current gate noise
    noise_chan_applier_sequence_this_gate = NoiseChannelApplierSequense()
    if gate_noise is not None:
        if type(gate_noise) is NoiseChannel:
            noise_chan_applier_sequence_this_gate.add(noise_chan=NoiseChannelSequence(gate_noise), qbit_list=qubit_list)            
        elif type(gate_noise) is NoiseChannelSequence:
            noise_chan_applier_sequence_this_gate.add(noise_chan=gate_noise, qbit_list=qubit_list)
        else:
            raise qckt.QCktException('ERROR: NoiseChannel, NoiseChannelSequence object or None expected.')
    noise_chan_applier_sequence_overall.extend(noise_chan_applier_sequence_this_gate)

    return noise_chan_applier_sequence_overall



#!/usr/bin/env python3

import sys
import numpy as np
import random as rnd
import qckt.gatesutils as gutils
import qckt.noisemodel as ns
from qckt.qException import QCktException

class QGate:
	def __init__(self):
		self.qbits = None
		self.cbits = None
		self.gateparams = []
		self.name = None
		self.opMatrix = None
		self.cbit_cond = None
		self.noise_chan = None

	#
	# Realign qubits
	#
	def _reorderlist(self,oseq,nseq):
		newseq = []
		nqbits = len(nseq)
		for i in oseq:
			if issubclass(type(i),list):
				newseq.append(self._reorderlist(i,nseq))
			else:
				newseq.append(nseq[nqbits-i-1])
		return newseq

	def realign(self,newseq):
		self.qbits = self._reorderlist(self.qbits,newseq)
		if self.cbits is not None:
			self.cbits = self._reorderlist(self.cbits,newseq)
		return self

	#
	# update opmatrix for multiple qubits
	#
	def __check_and_update_opmatrix_for_qubitlist__(self):
		ret_opMatrix = self.opMatrix
		ret_qbits = self.qbits
		if issubclass(type(self.qbits[0]),list):
			# combine_par to create opMatrix to convert from 1-qubit to multiple qubits
			opMatList = [self.opMatrix for _ in range(len(self.qbits[0]))]
			ret_opMatrix = gutils.combine_opmatrices_par(op_list=opMatList)
			ret_qbits = self.qbits[0]
			# print(f'gate {self.name} shape {opMatrix.shape} qubits are a list = {qbits}', file=sys.stderr)
		return ret_opMatrix, ret_qbits

	#
	# assemble into normalized form 
	#
	def assemble(self, noise_profile, noise_profile_gates):
		noise_chan = self.get_gate_noise(noise_profile_gates)
		noise_opseq = ns.consolidate_gate_noise(noise_profile=noise_profile, gate_noise=noise_chan, qubit_list=self.qbits)
		return {"op":"gate","gateObj":self,"name":self.name,"opMatrix":self.opMatrix,"qubits":self.qbits, 'ifcbit': self.cbit_cond, 'noiseChan':noise_opseq}

	#
	# convert opMatrix to full circuit nqubits size
	#
	def to_fullmatrix(self,nqbits):
		oplist = []
		if issubclass(type(self.qbits[0]),list) and len(self.qbits) == 1:
			for q in self.qbits[0]:
				op = gutils.stretched_opmatrix(nqbits,self.opMatrix,[q])
				oplist.append(op)
			opmat = gutils.combine_opmatrices_seq(oplist)
		else:
			opmat = gutils.stretched_opmatrix(nqbits,self.opMatrix,self.qbits)
		return opmat

	#
	# set conditional execution
	#
	def ifcbit(self,cbit,val):
		self.cbit_cond = (cbit,val)
		return self
	
	#
	# noise related functions
	#####################################################
	#
	# set noise channel for this gate instance
	#
	def set_noise(self, noise_chan):  # noise_chan can be NoiseChannel or NoiseChannelSequence
		# don't need to check NoiseChannel or NoiseChannelSequence, consolidate_gate_noise() checks that and converts
		if self.check_noise_chan_multi_qubits(noise_chan) == False:
			raise QCktException(f"Error: set_noise() - multi-qubit noise operator, {noise_chan.name},  can't be used on gate {self.name}")
		self.noise_chan = noise_chan
		return self
	
	#
	# validate noise chan and if multi-qubit noise chan applicable
	#
	def check_noise_chan_multi_qubits(self, noise_chan):
		if noise_chan is not None:
			if type(noise_chan) is not ns.NoiseChannel and type(noise_chan) is not ns.NoiseChannelSequence:
				raise QCktException(f"ERROR: NoiseChannel or NoiseChannelSequence object or None expected.")
			if type(noise_chan) is ns.NoiseChannel:
				if noise_chan.nqubits != 1 and len(self.qbits) != noise_chan.nqubits:
					return False
			else:  # type is ns.NoiseChannelSequence
				for op in noise_chan:
					if op.nqubits != 1 and len(self.qbits) != op.nqubits:
						return False
		return True

	#
	# pick from noise on gate type, and noise on gate instance
	#
	def get_gate_noise(self, noise_profile_gates):
		noise_chan = noise_profile_gates.get(self.__class__, None)
		if self.check_noise_chan_multi_qubits(noise_chan) == False:
			raise QCktException(f"Error: add_noise_to_all() - multi-qubit noise operator, {noise_chan.name},  can't be used on gate {self.name}")
		if self.noise_chan is not None:
			noise_chan = self.noise_chan
		return noise_chan
	
	#
	# if this gate instance is a step in circuit to apply noise step, overridden by indivivdual gates
	#
	def is_noise_step(self):  # is a step after which 'allsteps' noise should be applied
		return True

	#
	# circuit draw related functions
	#####################################################
	#
	# add this gate instance to canvas
	#
	def addtocanvas_gatenoise(self, canvas, noise_profile, noise_profile_gates,tag):
		noise_ch_appl_seq = self.form_gatenoise_ch_appl_seq(noise_profile=noise_profile, noise_profile_gates=noise_profile_gates)
		for kop,qbt in noise_ch_appl_seq:
			canvas._add_simple(qbt,f'{tag}:{kop.name}')

	# form a noise chan applier sequence for this gate instance
	# NOISE, M, Probe, Border override it to return an empty NoiseChannelApplierSequence()
	def form_gatenoise_ch_appl_seq(self, noise_profile, noise_profile_gates):
		noise_chan = self.get_gate_noise(noise_profile_gates)
		noise_ch_appl_seq = ns.consolidate_gate_noise(noise_profile=noise_profile, gate_noise=noise_chan, qubit_list=self.qbits)
		return noise_ch_appl_seq

	##############################################################################
	## Args validation methods
	##############################################################################

	def check_qbit_args(self,nqbits):
		## to be overridden by individual gates
		print("TODO: Args validation missing for ",type(self).__name__)

	def oneqbit_args(self,nqbits,qbits_list=None):
		multiqbits = []
		retval = None
		qbits = self.qbits
		if qbits_list is not None:
			qbits = qbits_list
		if len(qbits) == 1:
			if type(qbits[0]) is int:
				multiqbits = self.qbits
				retval = None # just asserting that retval is None here
			elif issubclass(type(qbits[0]),list):
				if len(qbits[0]) > 0:
					retval = None # just asserting that retval is None here
					for q in qbits[0]:
						if type(q) is int:
							multiqbits.append(q)
						else:
							retval = False
				else:
					retval = False
			else:
				retval = False
		else:
			retval = False
		if retval is None:
			retval = self.isvalid_qbits_gen(nqbits,multiqbits)
		return retval

	def fixedqbit_args(self,nqbits,nfixedq,qbits_list=None):
		multiqbits = []
		retval = None
		qbits = self.qbits
		if qbits_list is not None:
			qbits = qbits_list
		if len(qbits) == nfixedq:
			for q in qbits:
				if type(q) is int:
					multiqbits.append(q)
				else:
					retval = False
		else:
			retval = False
		if retval is None:
			retval = self.isvalid_qbits_gen(nqbits,multiqbits)
		return retval

	def varqbit_args(self, nqbits, minnq, qbits_list=None):
		multiqbits = []
		retval = None
		qbits = self.qbits
		if qbits_list is not None:
			qbits = qbits_list
		if len(qbits) >= minnq:
			for q in qbits:
				if type(q) is int:
					multiqbits.append(q)
				else:
					retval = False
		else:
			retval = False
		if retval is None:
			retval = self.isvalid_qbits_gen(nqbits, multiqbits)
		return retval

	def isvalid_qbits_gen(self,nqbits,qbit_list):
		if len(qbit_list) > nqbits:
			return False
		for i in qbit_list:
			if i >= nqbits:
				return False
			if qbit_list.count(i) != 1:
				return False
		return True

# Decorator to regiter the gate class in the GatesList list
GatesList = []
def registerGate(gateclass):
	GatesList.append(gateclass)
	return gateclass

def deregisterGate(gate_name):
	for gclass in GatesList:
		if gclass.__name__ == gate_name:
			GatesList.remove(gclass)

def get_gates_list():
	gates_list = []
	for gclass in GatesList:
		gates_list.append(gclass.__name__)
	return gates_list

@registerGate
class NOISE(QGate):
	def __init__(self, noise_chan, qbit):
		super().__init__()
		self.qbits = qbit
		if type(noise_chan) is ns.NoiseChannel:
			noise_chan = ns.NoiseChannelSequence(noise_chan)
		self.noise_chan = noise_chan
		self.name = self.noise_chan.name

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits, f'{tag}:{self.name}')

	def form_gatenoise_ch_appl_seq(self, noise_profile, noise_profile_gates):
		return ns.NoiseChannelApplierSequense()

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	def assemble(self, noise_profile, noise_profile_gates):
		noise_chan = self.noise_chan
		if type(noise_chan) is ns.NoiseChannel:
			print('Internal warning: found NOISE noise_chan of class NoiseChannel', file=sys.stderr)
			noise_chan = ns.NoiseChannelSequence(self.noise_chan)
		return {"op":"noise","gateObj":self,"name":self.name,"noiseChan":noise_chan,"qubits":self.qbits}

	def to_fullmatrix(self,nqbits):
		return None

	def is_noise_step(self):
		return False

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class X(QGate):
	def __init__(self, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.name = "X"
		self.opMatrix = np.matrix([[0,1],[1,0]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()
	
	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"X", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class Y(QGate):
	def __init__(self, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.name = "Y"
		self.opMatrix = np.matrix([[0,complex(0,-1)],[complex(0,1),0]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()
	
	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"Y", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class Z(QGate):
	def __init__(self, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.name = "Z"
		self.opMatrix = np.matrix([[1,0],[0,-1]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()
	
	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"Z", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class H(QGate):
	def __init__(self, qbit):
		super().__init__()
		sqr2 = np.sqrt(2)
		self.qbits = [qbit]
		self.name = "H"
		self.opMatrix = np.matrix([[1/sqr2,1/sqr2],[1/sqr2,-1/sqr2]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()
	
	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"H", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CX(QGate):

	def __init__(self, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.name = "CX"
		self.opMatrix = np.matrix([[0,1],[1,0]],dtype=complex)
		for i in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["X"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CY(QGate):

	def __init__(self, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.name = "CY"
		self.opMatrix = np.matrix([[0,complex(0,-1)],[complex(0,1),0]],dtype=complex)
		for i in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["Y"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CZ(QGate):

	def __init__(self, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.name = "CZ"
		self.opMatrix = np.matrix([[1,0],[0,-1]],dtype=complex)
		for i in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["Z"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):


@registerGate
class CCX(QGate):

	def __init__(self, control1, control2, target):
		super().__init__()
		self.qbits = [control1, control2, target]
		self.name = "CCX"
		self.opMatrix = gutils.CTL(gutils.CTL(np.matrix([[0,1],[1,0]],dtype=complex)))

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,[".",".","X"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.fixedqbit_args(nqbits,3)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class SWAP(QGate):

	def __init__(self, qbit1, qbit2):
		super().__init__()
		self.qbits = [qbit1, qbit2]
		self.name = "SWAP"
		self.opMatrix = np.matrix([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]],dtype=complex)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["*","*"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.fixedqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class M(QGate):

	def __init__(self, qubitslist, clbitslist=None):
		super().__init__()
		if clbitslist is None:
			clbitslist = qubitslist
		if (not issubclass(type(qubitslist),list)) or (not issubclass(type(clbitslist),list)):
			errmsg = "Error: M gate requires a list of qubits and optionally a list of classical bits as arguments"
			raise QCktException(errmsg)
		self.qbits = qubitslist
		self.cbits = clbitslist
		self.name = "M"

	def addtocanvas(self,canvas,tag):
		col = canvas._get1col(3)
		en = len(col)//2 - 1
		for qb in sorted(self.qbits):
			st = qb
			for i in range(st,en):
				col[2*i] = "-|-"
				col[2*i+1] = " | "
			col[qb*2] = "[M]"
		col[en*2] = "=v="
		canvas._append(col)
		canvas._extend()
		return self

	def form_gatenoise_ch_appl_seq(self, noise_profile, noise_profile_gates):
		return ns.NoiseChannelApplierSequense()

	def check_qbit_args(self,nqbits):
		retval = True
		if len(self.qbits) != len(self.cbits):
			retval = False
		if not self.varqbit_args(nqbits,1):
			retval = False
		if not self.varqbit_args(nqbits,1,qbits_list=self.cbits):
			retval = False
		return retval

	def ifcbit(self,cbit,val):
		raise QCktException("M gate does not support ifcbit()")

	# INHERIT realign(self,newseq):
	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

	def to_fullmatrix(self,nqbits):
		errmsg = "Measure gate cannot be converted to opmatrix."
		raise QCktException(errmsg)

	def assemble(self, noise_profile, noise_profile_gates):
		return {"op":"measure", "gateObj":self, "name":self.name, "qubits":self.qbits, "clbits":self.cbits}

	def is_noise_step(self):
		return False

@registerGate
class Border(QGate):

	def __init__(self):
		super().__init__()
		self.qbits = []
		self.name = "BORDER"

	def addtocanvas(self,canvas,tag):
		canvas._extend()
		col = canvas._get1col(1)
		en = len(col)//2
		for i in range(en):
			col[2*i] = "#"
			col[2*i+1] = "#"
		canvas._append(col)
		canvas._extend()
		return self

	def form_gatenoise_ch_appl_seq(self, noise_profile, noise_profile_gates):
		return ns.NoiseChannelApplierSequense()

	def check_qbit_args(self,nqbits):
		pass

	def ifcbit(self,cbit,val):
		raise QCktException("Border does not support ifcbit()")

	# INHERIT realign(self,newseq):

	def to_fullmatrix(self,nqbits):
		return None

	def assemble(self, noise_profile, noise_profile_gates):
		return {"op":"noop","gateObj":self}

	def is_noise_step(self):
		return False

	# OVERRIDE __str__(self) - no [qbits]
	def __str__(self):
		return "BORDER"

@registerGate
class Probe(QGate):

	def __init__(self, header="",probestates=None,probeobject=None):
		super().__init__()
		self.qbits = []
		self.header = "PROBE: "+header
		self.probestates = probestates
		self.probeobject = probeobject
		self.name = "Probe"

	def addtocanvas(self,canvas,tag):
		canvas._extend()
		col = canvas._get1col(1)
		en = len(col)//2
		for i in range(en):
			col[2*i] = "v"
			col[2*i+1] = "v"
		canvas._append(col)
		canvas._extend()
		return self

	def form_gatenoise_ch_appl_seq(self, noise_profile, noise_profile_gates):
		return ns.NoiseChannelApplierSequense()

	def check_qbit_args(self,nqbits):
		pass

	def ifcbit(self,cbit,val):
		raise QCktException("Probe does not support ifcbit()")

	# INHERIT realign(self,newseq):

	def to_fullmatrix(self,nqbits):
		return None

	def assemble(self, noise_profile, noise_profile_gates):
		return {"op":"probe","gateObj":self,"header":self.header,"probestates":self.probestates}

	def is_noise_step(self):
		return False

	# OVERRIDE __str__(self) - no [qbits]
	def __str__(self):
		return "PROBE"

class ProbeAnalyzer:
	'if an object of ProbeAnalyzer class is passed in Probe gate, its analyze() method is called when Probe gate is invoked.'
	def __init__(self):
		pass
	def analyze(self,creglist,sveclist):
		pass


@registerGate
class QFT(QGate):

	def __init__(self, *allqbits):
		super().__init__()
		if len(allqbits) == 1 and issubclass(type(allqbits[0]),list):
			self.qbits = allqbits[0]
		else:
			self.qbits = list(allqbits)
		self.name = "QFT"

		N = 2**len(self.qbits) # number of rows and cols
		theta = 2.0 * np.pi / N
		opMat = [None]*N
		for i in range(N):
			row = []
			for j in range(N):
				pow = i * j
				pow = pow % N
				# print "w^",pow
				row.append(np.e**(1.j*theta*pow))
			opMat[i] = row
		self.opMatrix = np.matrix(opMat,dtype=complex) / np.sqrt(N)

	def addtocanvas(self,canvas,tag):
		canvas._add_boxed(self.qbits,"QFT", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class RND(QGate):
	def __init__(self, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.name = "RND"

		phi = rnd.random() * np.pi * 2
		camp = np.cos(phi)
		samp = np.sin(phi)
		phi = rnd.random() * np.pi * 2
		re1 = camp*np.cos(phi)
		im1 = camp*np.sin(phi)
		re2 = samp*np.cos(phi)
		im2 = samp*np.sin(phi)
		self.opMatrix = np.matrix([[complex(-re1,-im1),complex(re2,im2)],[complex(re2,im2),complex(re1,im1)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"RND", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class P(QGate):
	def __init__(self, rotphi, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.gateparams = [rotphi]
		self.name = "P"

		phi = self.gateparams[0]
		cphi = np.cos(phi)
		sphi = np.sin(phi)
		self.opMatrix = np.matrix([[1,0],[0,complex(cphi,sphi)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"P", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CP(QGate):

	def __init__(self, rotphi, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.gateparams = [rotphi]
		self.name = "CP"

		phi = self.gateparams[0]
		cphi = np.cos(phi)
		sphi = np.sin(phi)
		self.opMatrix = np.matrix([[1,0],[0,complex(cphi,sphi)]],dtype=complex)
		for q in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["P"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class UROTk(QGate):
	def __init__(self, rotk, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.gateparams = [rotk]
		self.name = "UROTk"

		k = self.gateparams[0]
		ck = np.cos(2*np.pi/(2**k))
		sk = np.sin(2*np.pi/(2**k))
		self.opMatrix = np.matrix([ [1,0], [0,complex(ck,sk)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"UROTk", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CROTk(QGate):
	def __init__(self, rotk, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.gateparams = [rotk]
		self.name = "CROTk"

		k = self.gateparams[0]
		ck = np.cos(2*np.pi/(2**k))
		sk = np.sin(2*np.pi/(2**k))
		self.opMatrix = np.matrix([ [1,0], [0,complex(ck,sk)]],dtype=complex)
		for i in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["UROTk"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class Rx(QGate):
	def __init__(self, theta, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.gateparams = [theta]
		self.name = "Rx"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,0),complex(0,-sk)], [complex(0,-sk),complex(ck,0)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"Rx", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CRx(QGate):
	def __init__(self, theta, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.gateparams = [theta]
		self.name = "CRx"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,0),complex(0,-sk)], [complex(0,-sk),complex(ck,0)]],dtype=complex)
		for q in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["Rx"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):


@registerGate
class Ry(QGate):
	def __init__(self, theta, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.gateparams = [theta]
		self.name = "Ry"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,0),complex(-sk,0)], [complex(sk,0),complex(ck,0)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"Ry", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CRy(QGate):
	def __init__(self, theta, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.gateparams = [theta]
		self.name = "CRy"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,0),complex(-sk,0)], [complex(sk,0),complex(ck,0)]],dtype=complex)
		for q in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["Ry"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):


@registerGate
class Rz(QGate):
	def __init__(self, theta, qbit):
		super().__init__()
		self.qbits = [qbit]
		self.gateparams = [theta]
		self.name = "Rz"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,-sk),complex(0,0)], [complex(0,0),complex(ck,sk)]],dtype=complex)
		self.opMatrix, self.qbits = self.__check_and_update_opmatrix_for_qubitlist__()

	def addtocanvas(self,canvas,tag):
		canvas._add_simple(self.qbits,"Rz", self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,1)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

@registerGate
class CRz(QGate):
	def __init__(self, theta, *allqbits):
		super().__init__()
		self.qbits = list(allqbits)
		self.gateparams = [theta]
		self.name = "CRz"

		rot = self.gateparams[0]
		ck = np.cos(rot/2)
		sk = np.sin(rot/2)
		self.opMatrix = np.matrix([ [complex(ck,-sk),complex(0,0)], [complex(0,0),complex(ck,sk)]],dtype=complex)
		for q in range(len(self.qbits)-1):
			self.opMatrix = gutils.CTL(self.opMatrix)

	def addtocanvas(self,canvas,tag):
		canvas._add_connected(self.qbits,["."]*(len(self.qbits)-1)+["Rz"], self.cbit_cond)
		return self

	def check_qbit_args(self,nqbits):
		return self.varqbit_args(nqbits,2)

	# INHERIT def realign(self,newseq):
	# INHERIT def __str__(self):

def define_gate(gate_name, opMatrix):
	if not gutils.isunitary(opMatrix):
		errmsg = f"Custom gate, {gate_name}, operator matrix is not unitary."
		raise QCktException(errmsg)
	if gate_name in get_gates_list():
		print(f'WARNING: Gate {gate_name} already exists, oevrriding definition.', file=sys.stderr)
		deregisterGate(gate_name=gate_name)
	
	@registerGate
	class USER_DEFINED_GATE(QGate):
		def __init__(self, *qbits):
			super().__init__()
			self.name = gate_name
			self.opMatrix = opMatrix
			self.qbits = [q for q in qbits]
		def addtocanvas(self,canvas,tag):
			canvas._add_boxed(self.qbits,self.name, self.cbit_cond)
			return self
		def check_qbit_args(self,nqbits):
			# print('check_qbit_args called')
			r,c = self.opMatrix.shape
			if 2**len(self.qbits) != r:
				return False
			return self.varqbit_args(nqbits,1)
		# INHERIT def realign(self,newseq):
		# INHERIT def __str__(self):
	USER_DEFINED_GATE.__name__ = gate_name



#!/usr/bin/python3

import qckt.backend.QSystems as qsys
from qckt.qException import QCktException


class Registry:
	'this is an API for accessing the registry of Quantum Computing Services (e.g., IBM quantum computing, Ionq quantum computing, local qsim simulator, local debugging simulator)'
	services = {"QSystems": (qsys.qsimSvc, "qsim simulator from Quantum Systems")}

	def __init__(self):
		pass
	def listSvc(self):
		'returns a list of tuples (name, description) of all services available (i.e. registered in the configuration)'
		return [(k,Registry.services[k][1]) for k in Registry.services.keys()]
	def getSvc(self, svcName):
		if svcName in Registry.services.keys():
			return Registry.services[svcName][0]()
		else:
			raise QCktException("No such registered service: "+str(svcName))




import sys

class Job:
	'just a packet of all job attributes. Is a qckt concept, so is '
	def __init__(self, circuit, qtrace=False, verbose=False, shots=1):
		'Job is a qckt concept to encapsulate the work packet that is to be submitted to the quantum computer to execute'
		self.circuit = circuit
		self.nqubits, self.nclbits = circuit.get_size()
		self.assembledCkt = circuit.assemble()
		self.qtrace = qtrace
		self.verbose = verbose
		self.shots = shots
		# result will be populated after the job is run
		self.result = None
		self.runstats = None

	def get_svec(self):
		return self.result.state_vector

	def get_creg(self):
		return self.result.cregister

	def get_counts(self, register=None):
		if register is not None:
			# register specifies which classical register to get counts for
			# for now, only single classical register is supported
			for bit in register:
				if bit < 0 or bit >= self.nclbits:
					raise ValueError("Invalid classical register specified")
				if sum([1 if bit == b else 0 for b in register]) > 1:
					raise ValueError("Duplicate bits in classical register specified")
			mask = 0
			for bit in register:
				mask |= (1 << bit)
			counts = {}
			for i in sorted(self.result.creg_counts.keys()):
				c = self.result.creg_counts[i]
				state = i & mask
				counts[state] = counts.get(state,0) + c
			return counts
		else:
			# just to have the keys in a sorted order
			counts = {}
			for i in sorted(self.result.creg_counts.keys()):
				counts[i] = self.result.creg_counts[i]
			return counts

	def plot_counts(self,register=None, verbose=False):
		counts = self.get_counts(register=register)
		if 'ipykernel' in sys.modules:
			lbl = []
			vals = []
			for i in range(2**self.nqubits):
				c = counts.get(i,0)
				if verbose or c > 0:
					binlbl = ('[{0:0'+str(self.nqubits)+'b}]').format(i)
					lbl.append(str(i)+" "+binlbl)
					vals.append(c)
			import matplotlib.pyplot as plt
			plt.xticks(rotation='vertical')
			fig = plt.bar(lbl,vals)
			for i in range(len(vals)):
				plt.annotate(str(vals[i]), xy=(lbl[i],vals[i]), ha='center', va='bottom')
			plt.show()
		else:
			maxc = max([val for k,val in counts.items()])
			cwid = len(str(maxc))
			scale = 1 if maxc < 50 else 50.0/maxc
			for i in range(2**self.nqubits):
				c = counts.get(i,0)
				if verbose or c > 0:
					print(('{0:4d} [{0:0'+str(self.nqubits)+'b}]  {1:>'+str(cwid)+'d} |').format(i,c),end="")
					for j in range(int(c * scale)):
						print("*",end="")
					print()
		return counts

	def get_runstats(self):
		return self.runstats

	def print_runstats(self):
		qsteps = self.runstats['QSteps']
		op_counts = self.runstats['OpCounts']
		op_times = self.runstats['OpTimes']
		tot_time = 0.0
		for k,v in op_times.items():
			tot_time = tot_time + v
		print(f'Total Ops  :  {qsteps:4d}      operations')
		# print(f'Total Time :  {tot_time:9.4f} sec')
		print(f'Per Operation:')
		for op in op_counts.keys():
			# print(f'    {op:8s} {op_times[op]:9.4f} sec  {op_counts[op]:4d} times {op_times[op]/op_counts[op]:4.4f} avg')
			print(f'    {op:8s} {op_counts[op]:4d} times')




#!/usr/bin/env python

import numpy as np

# ABSTRACT class for backend services
class BackendSvc:
	'the service implments the methods to connect with the quantum computing service, i.e., the required protocol, etc. - config registry will provide the specifics e.g., IP address'
	def __init__(self,connectionToken=None):
		'connectionToken contains the credentials required to access the service; the type of this object is specific to the service'
		self.connnectionToken = connectionToken
		pass
	def listInstances(self):
		'provides a list of tuples (name, description) of all instances (quamtum computer) available at this service'
		pass
	def getInstance(self, name):
		'returns an object representation of the named instance(quantum computer)'
		pass

class Cregister:

	def __init__(self):
		self.value = None
		self.intvalue = None

	def __str__(self):
		creg_str = ""
		for i in (range(len(self.value))): # creg[0] is MSB
			creg_str = creg_str + "{0:01b}".format(self.value[i])
		# creg_str = creg_str + "\n"
		return creg_str

	def setvalue_vec(self,cbitsvec):
		intval = 0
		for i in range(len(cbitsvec)):
			intval = 2 * intval
			intval += cbitsvec[i]
		self.intvalue = intval
		self.value = cbitsvec


class StateVector:

	def __init__(self):
		self.value = None
		self.verbosity = False

	def __str__(self):
		svec_str = ""
		dispvalue = self.value
		if len(dispvalue.shape) > 1:
			svec_str = "Density matrix DIAGONAL\n"
			dispvalue = np.diagonal(self.value)

		### wonky way to find number of qubits
		nq = 0
		n = len(dispvalue) - 1
		while n > 0:
			nq = nq + 1
			n = n >> 1

		statefmt = "{0:"+"0{0:d}b".format(nq)+"}"
		s = 0
		for i in dispvalue:
			if self.verbosity or np.absolute(i) > 10**(-6):
				svec_str = svec_str + statefmt.format(s) + "  " + "{:.8f}".format(i) + "\n"
			s += 1
		return svec_str

	def verbose(self,v):
		if v is True:
			self.verbosity = True
		else:
			self.verbosity = False
		return self

class Result:
	def __init__(self,cregvals,svecvals=None, cregcounts=None):
		self.cregister = cregvals
		self.state_vector = svecvals
		self.creg_counts = cregcounts




import math
import qckt.gatesutils as gutils
from qckt.noisemodel.noiseUtils import NoiseChannel
import numpy as np
from qsim import QSimError


I = ['I',np.matrix([[1.0,0.0],[0.0,1.0]],dtype=complex)]
X = ['X', np.matrix([[0,1],[1,0]],dtype=complex)]
Y = ['Y', np.matrix([[0,complex(0,-1)],[complex(0,1),0]],dtype=complex)]
Z = ['Z', np.matrix([[1,0],[0,-1]],dtype=complex)]
I2 = ['I2', np.matrix([[1.0,0.0,0.0,0.0],[0.0,1.0,0.0,0.0],[0.0,0.0,1.0,0.0],[0.0,0.0,0.0,1.0]],dtype=complex)]

# Canned noise channels
def bit_flip(probability=0.05):
    noise_channel = NoiseChannel(
        name=f'BF({probability:.2f})',
        noise_chan=[(X,probability),(I,(1-probability))],
        nqubits=1,
        )
    return noise_channel


def phase_flip(probability=0.05):
    noise_channel = NoiseChannel(
        name=f'PF({probability:.2f})',
        noise_chan=[(Z,probability),(I,(1-probability))],
        nqubits=1,
        )
    return noise_channel


def depolarizing(probability=0.05):
    noise_channel = NoiseChannel(
        name=f'Dep({probability:.2f})',
        noise_chan=[(X,probability/3.0),(Y,probability/3.0),(Z,probability/3.0),(I,(1-probability))],
        nqubits=1,
        )
    return noise_channel


def amplitude_damping(gamma=0.05):
    AD_K1 = ['K1',np.matrix([[1.0,0.0],[0.0,math.sqrt(1-gamma)]], dtype=complex)]
    AD_K2 = ['K2',np.matrix([[0.0,math.sqrt(gamma)],[0.0,0.0]], dtype=complex)]
    noise_channel = NoiseChannel(
        name=f'AD({gamma:.2f})',
        noise_chan=[(AD_K1,1.0),(AD_K2,1.0)],
        nqubits=1,
        )
    return noise_channel


def phase_damping(gamma=0.05):
    PD_K0 = ['K0',np.matrix([[1.0,0.0],[0.0,math.sqrt(1-gamma)]], dtype=complex)]
    PD_K1 = ['K1',np.matrix([[0.0,0.0],[0.0,math.sqrt(gamma)]], dtype=complex)]
    noise_channel = NoiseChannel(
        name=f'PD({gamma:.2f})',
        noise_chan=[(PD_K0,1.0),(PD_K1,1.0)],
        nqubits=1,
        )
    return noise_channel


def pauli_channel(probx=0.05, proby=0.05, probz=0.05):
    noise_channel = NoiseChannel(
        name=f'PC({probx:.2f},{proby:.2f},{probz:.2f})',
        noise_chan=[(X,probx),(Y,proby),(Z,probz),(I,(1-probx-proby-probz))],
        nqubits=1,
        )
    return noise_channel


def generalized_amplitude_damping(probability=0.05, gamma=0.05):
    GAD_K0 = ['K0',math.sqrt(probability)*np.matrix([[1.0,0.0],[0.0,math.sqrt(1-gamma)]], dtype=complex)]
    GAD_K1 = ['K1',math.sqrt(probability)*np.matrix([[0.0,math.sqrt(gamma)],[0.0,0.0]], dtype=complex)]
    GAD_K2 = ['K2',math.sqrt(1-probability)*np.matrix([[math.sqrt(1-gamma),0.0],[0.0,1.0]], dtype=complex)]
    GAD_K3 = ['K3',math.sqrt(1-probability)*np.matrix([[0.0,0.0],[math.sqrt(gamma),0.0]], dtype=complex)]
    noise_channel = NoiseChannel(
        name=f'GAD({probability:.2f},{gamma:.2f})',
        noise_chan=[(GAD_K0,1.0),(GAD_K1,1.0),(GAD_K2,1.0),(GAD_K3,1.0)],
        nqubits=1,
        )
    return noise_channel
    # return {'name': f'GAD({probability:.2f},{gamma:.2f})', 'operator':[[(GAD_K0,1.0),(GAD_K1,1.0),(GAD_K2,1.0),(GAD_K3,1.0)], 0.0]}

def dummy_2qubit_chan(probability=0.05):
    noise_channel = NoiseChannel(
        name=f'DUMMY-CHAN({probability})',
        noise_chan=[(I2,probability),(I2,(1-probability))],
        nqubits=2,
        )
    return noise_channel

def two_qubit_dephasing(probability=0.05):
    zop = Z[1]
    iop = I[1]
    IZ = ['IZ',gutils.combine_opmatrices_par([iop,zop])]
    ZI = ['ZI',gutils.combine_opmatrices_par([zop,iop])]
    ZZ = ['ZZ',gutils.combine_opmatrices_par([zop,zop])]
    noise_chan = [
        (I2, (1-probability)),
        (IZ,probability/3.0),
        (ZI,probability/3.0),
        (ZZ,probability/3.0),
        ]
    noise_channel = NoiseChannel(
        name=f'2DPH({probability})',
        noise_chan=noise_chan,
        nqubits=2,
        )
    return noise_channel

def two_qubit_depolarizing(probability=0.05):
    noise_chan = [(I2,(1-probability))]
    for op1 in [I,X,Y,Z]:
        for op2 in [I,X,Y,Z]:
            if op1[0] == 'I' and op2[0] == 'I':
                continue
            opname = f'{op1[0]}{op2[0]}'
            opmatrix = gutils.combine_opmatrices_par([op1[1],op2[1]])
            twoqbitop = [opname,opmatrix]
            noise_chan.append((twoqbitop, (probability/15.0)))
    noise_channel = NoiseChannel(
        name=f'2DEP({probability})',
        noise_chan=noise_chan,
        nqubits=2,
    )
    return noise_channel


def noise_channel_list():
    return noise_chan_signature

def noise_channel_lookup(noise_channel_name):
    if noise_channel_name not in noise_channels.keys():
        raise QSimError(f'{noise_channel_name} not a valid pre-prepared noise channel.')
    return noise_channels[noise_channel_name]


noise_channels = {
    'BitFlip': bit_flip,
    'PhaseFlip': phase_flip,
    'Depolarizing': depolarizing,
    'AmplitudeDamping': amplitude_damping,
    'GeneralizedAmplitudeDamping': generalized_amplitude_damping,
    'PhaseDamping': phase_damping,
    'PauliChannel': pauli_channel,
    'TwoQubitDephasing': two_qubit_dephasing,
    'TwoQubitDepolarizing': two_qubit_depolarizing,
    }

noise_chan_signature = {}
for krname in noise_channels:
    fn = noise_channels[krname]
    fn_name = fn.__name__
    argcount = fn.__code__.co_argcount
    defcount = len(fn.__defaults__)
    nondefcount = argcount - defcount
    signature = f'{fn_name}('
    sep=''
    for i in range(argcount):
        argdesc = f'{fn.__code__.co_varnames[i]}'
        signature += f'{sep}{argdesc}'
        sep = ', '
        if i >= nondefcount:
            signature += f'={fn.__defaults__[i-nondefcount]}'
    signature += ')'
    noise_chan_signature[krname] = signature





from qckt.QCkt import QCkt
from qckt.Gates import define_gate, get_gates_list
from qckt.Job import Job
from qckt.qException import QCktException

__version__ = '3.1'
__author__ = 'Atul Varshneya'
__author_email__ = 'atul.varshneya@gmail.com'
__email__ = __author_email__
__url__ = 'https://github.com/atulvarshneya/quantum-computing'



#!/usr/bin/env python3

import qckt.Canvas as cnv
import qckt.Gates as gts
import qckt.gatesutils as gutils
import qckt.noisemodel as ns
from qckt.qException import QCktException
import numpy as np


class GateWrapper:
	def __init__(self, qckt, gateCls):
		self.qckt = qckt
		self.GateCls = gateCls
	def __call__(self, *args, **kwargs):
		gateObj = self.GateCls(*args, **kwargs)
		self.qckt.circuit.append(gateObj)
		if gateObj.check_qbit_args(self.qckt.nqubits) == False:
			raise QCktException(f"Error: qubit arguments incorrect. {gateObj.name}{str(gateObj.qbits)}")
		return gateObj
	def set_noise_on_all(self, noise_chan):
		# don't need to check NoiseChannel or NoiseChannelSequence, consolidate_gate_noise() checks that and converts
		self.qckt.noise_profile_gates[self.GateCls] = noise_chan


class QCkt:

	def __init__(self, nqubits, nclbits=0, name=None, noise_profile=None):
		self.nqubits = nqubits
		self.nclbits = nclbits
		self.circuit = []
		self.name = name
		self.canvas = cnv.Canvas(self)
		self.noise_profile_gates = {}  # this stores a map with gateCls as key, and noise spec as value
		self.noise_profile = None  # set_noise_profile() below updates it
		self.set_noise_profile(noise_profile)

		for gclass in gts.GatesList:
			setattr(self, gclass.__name__, GateWrapper(self,gclass))

		self.idx = 0 # for iterations

	def __iter__(self):
		self.it = iter(self.circuit)
		return self
	def __next__(self):
		return next(self.it)

	def get_gates_list(self):
		raise QCktException('ERROR: get_gates_list() is available under qckt package. See README.md. Usage: qckt.get_gates_list()')

	def append(self, otherckt, name=None):
		# otherckt - the ckt to be appended
		nq = max(self.nqubits, otherckt.nqubits)
		nc = max(self.nclbits, otherckt.nclbits)
		if name is None:
			name = self.name
		newckt = QCkt(nq,nc,name=name)

		# copy over noise_profile_gates list into new circuit
		newckt.__copyover_noise_profile_gates__(srcckt=self)
		newckt.__copyover_noise_profile_gates__(srcckt=otherckt)

		# copy circuits over to the new circuit
		for g in self.circuit:
			newckt.circuit.append(g)
		for g in otherckt.circuit:
			newckt.circuit.append(g)

		return newckt

	# change the qubits order to a different order
	def realign(self, newnq, newnc, inpqubits, name=None):
		# newq and newc are the new sizes of the qubits register and clbits register
		# inpqubits gives the new positions of the qubits
		# See README.md for details
		if self.nqubits != len(inpqubits):
			errmsg = "Error: error aligning qubits, number of qubits do not match"
			raise QCktException(errmsg)
		if name is None:
			name = self.name
		newckt = QCkt(newnq, newnc, name=name)

		# copy over noise_profile_gates list into new circuit
		newckt.__copyover_noise_profile_gates__(srcckt=self)

		for g in self.circuit:
			galigned = g.realign(inpqubits)
			newckt.circuit.append(galigned)
		return newckt

	def __copyover_noise_profile_gates__(self, srcckt):
		# copy over noise_profile_gates from source circuit
		for gcls in srcckt.noise_profile_gates:
			self.noise_profile_gates[gcls] = srcckt.noise_profile_gates[gcls]



	def assemble(self):
		assembled = []
		noise_wrapper = self.NOISE
		noise_cls = noise_wrapper.GateCls
		if self.noise_profile is not None and self.noise_profile.noise_chan_init is not None:
			noise_gate = noise_cls(self.noise_profile.noise_chan_init, list(range(self.nqubits)))
			cktstep = noise_gate.assemble(self.noise_profile, self.noise_profile_gates)
			cktstep['tag'] = 'INIT'
			assembled.append(cktstep)
		for gt in self.circuit:
			cktstep = gt.assemble(self.noise_profile, self.noise_profile_gates)
			cktstep['tag'] = 'NS' if cktstep['op'] == 'noise' else gt.name
			assembled.append(cktstep)
			if self.noise_profile is not None and self.noise_profile.noise_chan_allsteps is not None and gt.is_noise_step():
				for kop,qbt in self.noise_profile.noise_chan_allsteps:
					noise_gate = noise_cls(kop, qbt)
					cktstep = noise_gate.assemble(self.noise_profile, self.noise_profile_gates)
					cktstep['tag'] = 'AS'
					assembled.append(cktstep)
		return assembled

	def to_opMatrix(self):
		oplist = []
		for q in self.circuit:
			op = q.to_fullmatrix(self.nqubits)
			if op is not None: ## Border, Probe gates return None
				oplist.append(op)
		if len(oplist) == 0:
			return np.matrix(np.eye(2**self.nqubits),dtype=complex)
		opmat = gutils.combine_opmatrices_seq(oplist)
		return opmat

	def set_noise_profile(self, noise_profile=None):
		fixed_noise_profile = noise_profile
		if noise_profile is not None:
			# validate the keys in the dict for noise_profile argument
			if type(noise_profile) is not ns.NoiseProfile:
				raise QCktException(f'ERROR: noise_profile expected to be NoiseProfile object or None')

			# validate the noise_chan_init field
			opseq_init = noise_profile.noise_chan_init
			if opseq_init is not None:
				if type(opseq_init) is ns.NoiseChannelSequence:
					pass
				elif type(opseq_init) is ns.NoiseChannel:
					opseq_init = ns.NoiseChannelSequence(opseq_init)
				else:
					raise QCktException('ERROR: noise_profile.noise_chan_init must be NoiseChannelSequence or NoiseChannel object.')
				# if noise_chan_init is specified, it must have only 1-qubit noise channel
				for op in opseq_init:
					if op.nqubits != 1:
						raise QCktException(f'ERROR: noise_opseq_init must use 1-qubit noise channel')

			# validate the noise_chan_allgates field
			opseq_allgates = noise_profile.noise_chan_allgates
			if opseq_allgates is not None:
				if type(opseq_allgates) is ns.NoiseChannelSequence:
					pass
				elif type(opseq_allgates) is ns.NoiseChannel:
					opseq_allgates = ns.NoiseChannelSequence(opseq_allgates)
				else:
					raise QCktException('ERROR: noise_profile.noise_chan_allgates must be NoiseChannel or NoiseChannel object.')

			# validate the noise_chan_qubits field
			opseq_qubits = noise_profile.noise_chan_qubits
			if opseq_qubits is not None:
				if type(opseq_qubits) is not ns.NoiseChannelApplierSequense:
					raise QCktException('ERROR: noise_profile.noise_chan_qubits must be NoiseChannelApplierSequense object.')

			# validate the noise_chan_allsteps field
			opseq_allsteps = noise_profile.noise_chan_allsteps
			if opseq_allsteps is not None:
				if type(opseq_allsteps) is not ns.NoiseChannelApplierSequense:
					raise QCktException('ERROR: noise_profile.noise_chan_allsteps must be NoiseChannelApplierSequense object.')

			fixed_noise_profile = ns.NoiseProfile(noise_chan_init=opseq_init, noise_chan_allgates=opseq_allgates, noise_chan_qubits=opseq_qubits, noise_chan_allsteps=opseq_allsteps)
		# save the noise_profile
		self.noise_profile = fixed_noise_profile

	def get_size(self):
		return self.nqubits, self.nclbits

	def draw(self,show_noise=True):
		assembled_circuit = self.assemble()
		self.canvas.draw(assembled_circuit, show_noise=show_noise)

	def list(self, show_noise=True):
		assembled_circuit = self.assemble()
		if self.name is not None:
			print(self.name)
		for cktstep in assembled_circuit:
			if cktstep['op'] == 'gate':
				prline = f"{cktstep['name']}"
				gateparam_list = []
				for p in cktstep['gateObj'].gateparams:
					if type(p) is int:
						pstr = f'{p:d}'
					elif type(p) is float:
						pstr = f'{p:.4f}'
					else:
						pstr = str(p)
					gateparam_list.append(f'{pstr}')
				if len(gateparam_list) > 0:
					gateparam_str = ','.join(gateparam_list)
					prline = f'{prline}({gateparam_str})'
				prline = f"{prline}{cktstep['qubits']}"
				if cktstep['gateObj'].cbit_cond is not None:
					prline = f"{prline}.ifcbit{cktstep['gateObj'].cbit_cond}"
				print(prline)
				if show_noise:
					noise_ch_appl_seq = cktstep['gateObj'].form_gatenoise_ch_appl_seq(self.noise_profile, self.noise_profile_gates)
					if len(noise_ch_appl_seq.name) > 0:
						prline = f"{cktstep['name']}:{noise_ch_appl_seq.name}"
						print(f'* {prline}')
			elif cktstep['op'] == 'measure':
				prline = f"{cktstep['name']}"
				prline = f"{prline}{cktstep['qubits']}"
				if cktstep['gateObj'].cbits is not None:
					prline = f"{prline}{cktstep['gateObj'].cbits}"
				print(prline)
			elif cktstep['op'] == 'probe':
				prline = 'PROBE>>>>>>>>>>>>>>'
				print(prline)
			elif cktstep['op'] == 'noop':
				prline = 'BORDER#############'
				print(prline)
			elif cktstep['op'] == 'noise':
				if show_noise:
					prline = f"{cktstep['tag']}:{cktstep['name']}{cktstep['qubits']}"
					print(f'* {prline}')
			else:
				raise QCktException(f"INTERNAL ERROR: Unknown op {cktstep['op']}")

	def custom_gate(self, cgate_name, opMatrix):
		raise QCktException('ERROR: custom_gate() is discontinued, use qckt.define_gate() instead. See README.md.')


if __name__ == "__main__":
	pass



#!/usr/bin/python3

from qckt.qException import QCktException

class Register(list):
	def __init__(self,sz):
		super().__init__([None]*sz)
		self.placed = False
	def __getitem__(self,idx):
		if self.placed is False:
			raise QCktException("Attempt to access register qbit ids without placement")
		return super().__getitem__(idx)
	def placeMSB(self,loc):
		for q in range(len(self)):
			self[q] = loc - q
		self.placed = True
		return self
class QRegister(Register):
	pass
class CRegister(Register):
	pass

def placement(*regs):
	qbitplaced = QRegister(0).placeMSB(0) # trick to mark the register 'placed'
	cbitplaced = CRegister(0).placeMSB(0) # trick to mark the register 'placed'
	qbitsz = 0
	cbitsz = 0
	for r in reversed(regs):
		if type(r) is QRegister:
			if r.placed: raise QCktException("Attempt to place already placed QRegister")
			qbitsz = qbitsz + len(r) if qbitsz is not None else len(r)
			r.placeMSB(qbitsz - 1)
			temp = QRegister(0).placeMSB(0) # trick to mark the register 'placed'
			for i in r: temp.append(i)
			temp.extend(qbitplaced)
			qbitplaced = temp
		elif type(r) is CRegister:
			if r.placed: raise QCktException("Attempt to place already placed CRegister")
			cbitsz = cbitsz + len(r) if cbitsz is not None else len(r)
			r.placeMSB(cbitsz - 1)
			temp = CRegister(0).placeMSB(0) # trick to mark the register 'placed'
			for i in r: temp.append(i)
			temp.extend(cbitplaced)
			cbitplaced = temp
		else:
			raise QCktException("placement not supported for type:"+type(r).__name__)
	return qbitsz,cbitsz,qbitplaced,cbitplaced



#!/usr/bin/env python3

import qckt.Gates

class Canvas:

	def __init__(self,circuit):
		self.circuit = circuit
		(self.nqubits, self.nclbits) = self.circuit.get_size()
		self.paper = None
		self.lsb_on_top = True # default is to have LSB on top, MSB at the bottom

	def _initcanvas(self):
		self.paper = [[None]*(self.nqubits*2+2)]
		for i in range(self.nqubits+1):
			self.paper[0][i*2] = "q{0:03d} ".format(i)
			self.paper[0][i*2+1] = "     "
		self.paper[0][self.nqubits*2] = "creg "
		self._extend()
		return self

	def _get1col(self, width=1):
		col = [None]*(self.nqubits*2+2)
		for i in range(self.nqubits+1):
			col[i*2] = "-"*width
			col[i*2+1] = " "*width
		col[self.nqubits*2] = "="*width
		return col

	def _extend(self):
		col = self._get1col(1)
		self.paper.append(col)
		return self

	def _add_simple(self,qbit,symbol,ccond=None):
		if issubclass(type(qbit),list):
			return self._add_multiple(qbit,symbol,ccond)
		else:
			ncols = len(symbol) + 2
			col = self._get1col(ncols)
			col[2*qbit] = "[" + symbol + "]"
			if ccond is not None:
				st = qbit
				for i in range(st,self.nqubits):
					col[i*2+1] = ':' + col[i*2+1][1:]
				tag = f'{ccond[0]}'
				col[2*self.nqubits] = tag + col[2*self.nqubits][len(tag):]
			self._append(col)
			self._extend()
			return self

	def _add_multiple(self,qbits,symbol,ccond=None):
		ncols = len(symbol) + 2
		col = self._get1col(ncols)
		for i in range(len(qbits)):
			col[2*qbits[i]] = "[" + symbol + "]"
		if ccond is not None:
			st = min(qbits)
			for i in range(st,self.nqubits):
				col[i*2+1] = ':' + col[i*2+1][1:]
			tag = f'{ccond[0]}'
			col[2*self.nqubits] = tag + col[2*self.nqubits][len(tag):]
		self._append(col)
		self._extend()
		return self

	def _add_connected(self,qbits,symbols,ccond=None):
		ncols = max([len(e) + 2 for e in symbols])
		post = (ncols-1) // 2
		pre = (ncols-1) - post
		col = self._get1col(ncols)
		st = min(qbits)
		en = max(qbits)
		for i in range(st,en):
			col[i*2] = "-"*pre + "|" + "-"*post
			col[i*2+1] = " "*pre + "|" + " "*post
		for i in range(len(qbits)):
			padding = ncols - len(symbols[i]) - 2
			post = padding // 2
			pre = padding - post
			col[2*qbits[i]] = "-"*pre + "[" + symbols[i] + "]" + "-"*post
		if ccond is not None:
			for i in range(st,self.nqubits):
				col[i*2+1] = ':' + col[i*2+1][1:]
			tag = f'{ccond[0]}'
			col[2*self.nqubits] = tag + col[2*self.nqubits][len(tag):]
		self._append(col)
		self._extend()
		return self

	def _add_boxed(self,qbits,symbol,ccond=None):
		ncols =  len(symbol) + 4 # 4 to make space for "[", " M"/" L", "]"
		col = self._get1col(ncols)
		st = min(qbits)
		en = max(qbits)
		for i in range(st,en):
			col[i*2] = "|" + "-"*(ncols - 2) + "|"
			col[i*2+1] = "|" + " "*(ncols - 2) + "|"
		lsb = qbits[-1]
		msb = qbits[0]
		for q in qbits:
			if q == msb:
				col[2*msb] = "[" + symbol + " M]"
			elif q == lsb:
				col[2*lsb] = "[" + symbol + " L]"
			else:
				col[2*q] = "[" + symbol + "  ]"
		if ccond is not None:
			for i in range(en,self.nqubits):
				col[i*2+1] = ':' + col[i*2+1][1:]
			tag = f'{ccond[0]}'
			col[2*self.nqubits] = tag + col[2*self.nqubits][len(tag):]
		self._append(col)
		self._extend()
		return self

	def _paint(self):
		if self.circuit.name is not None:
			print(self.circuit.name)
		ncols = len(self.paper)
		nrows = len(self.paper[0])
		for i in range(nrows):
			for j in range(ncols):
				print(self.paper[j][i], end='')
			print()

	def _append(self,col):
		list = self.paper
		list.append(col)
		return self
	
	def draw(self,assembled_circuit,show_noise=True):
		self._initcanvas()
		for cktstep in assembled_circuit:
			if cktstep['op'] != 'noise':
				cktstep['gateObj'].addtocanvas(self,cktstep['tag'])
			elif show_noise:  # op is NOISE, and show_noise == True
				cktstep['gateObj'].addtocanvas(self,cktstep['tag'])
			if show_noise:
				cktstep['gateObj'].addtocanvas_gatenoise(self, noise_profile=self.circuit.noise_profile, noise_profile_gates=self.circuit.noise_profile_gates, tag=cktstep['tag'])
		self._paint()



#!/usr/bin/env python
import sys
import qsim
import qsim.noisemodel as nmdl
import qckt.noisemodel as ns
from qckt.backend.BackendAPI import *
from qckt.qException import QCktException

def convert_to_qsim_noise_op_applier_seq(qckt_noise_chan_applier_seq):
	qsim_noise_applier_seq = None
	if qckt_noise_chan_applier_seq is not None:
		qsim_noise_applier_seq = nmdl.qNoiseChannelApplierSequense()
		for chan,qbits in qckt_noise_chan_applier_seq:
			opname = chan.name
			noise_chan = chan.noise_chan
			nqubits = chan.nqubits
			qsim_op = nmdl.qNoiseChannel(name=opname,operator_prob_set=noise_chan,nqubits=nqubits)
			qsim_noise_applier_seq.add(noise_chan=qsim_op,qubit_list=qbits)
	return qsim_noise_applier_seq

def convert_to_qsim_noise_op_sequence(qckt_noise_chan_sequence):
	qsim_noise_opseq = None
	if qckt_noise_chan_sequence is not None:
		qsim_noise_opseq = nmdl.qNoiseChannelSequence()
		for chan in qckt_noise_chan_sequence:
			opname = chan.name
			opprobset = chan.noise_chan
			nqubits = chan.nqubits
			qsim_op = nmdl.qNoiseChannel(name=opname,operator_prob_set=opprobset,nqubits=nqubits)
			qsim_noise_opseq.add(qsim_op)
	return qsim_noise_opseq


class DMQeng:
	def __init__(self):
		pass

	def runjob(self, job):
		qc = qsim.DMQSimulator(job.nqubits,job.nclbits,noise_profile=None,qtrace=False,verbose=job.verbose)
		for op in job.assembledCkt:
			if op["op"] == "gate":
				qckt_noise_op = op['noiseChan']
				qsim_noise_opseq = convert_to_qsim_noise_op_applier_seq(qckt_noise_op)
				qc.qgate([op["name"],op["opMatrix"]], op["qubits"], ifcbit=op["ifcbit"],noise_chan=qsim_noise_opseq)
			elif op["op"] == "measure":
				qc.qmeasure(op["qubits"],cbit_list=op["clbits"])
			elif op["op"] == "noise":
				qckt_noise_op = op['noiseChan']
				qsim_noise_opseq = convert_to_qsim_noise_op_sequence(op["noiseChan"])
				qc.qnoise(qsim_noise_opseq, op['qubits'])
			elif op["op"] == "probe":
				pass
			elif op["op"] == "noop":
				pass
			else:
				raise QCktException("Encountered unknown instruction: "+str(op["op"]))
		# at the end of execution perform READOUT sampling the state as many times as job.shots
		qc.qreadout(nshots=job.shots)  # this will populate the creg_counts values internally, which are returned by qsnapshot()

		# Now create the Result object from the final snapshot and store it in job.result
		cregvec,statevecarr,cregcounts,runstats = qc.qsnapshot()
		cregres = Cregister()
		cregres.setvalue_vec(cregvec)
		# store the last statevec as well, helps with debugging
		statevec = StateVector()
		statevec.value = statevecarr
		job.result = Result(cregvals=cregres, cregcounts=cregcounts)
		job.runstats = runstats
		return self

class DMQdeb:
	def __init__(self):
		self.maxerr = 10**(-6)
		pass

	def runjob(self, job):
		qc = qsim.DMQSimulator(job.nqubits,job.nclbits,noise_profile=None,qtrace=job.qtrace,verbose=job.verbose)
		for op in job.assembledCkt:
			if op["op"] == "gate":
				qckt_noise_op = op['noiseChan']
				qsim_noise_opseq = convert_to_qsim_noise_op_applier_seq(qckt_noise_op)
				qc.qgate([op["name"],op["opMatrix"]], op["qubits"], ifcbit=op["ifcbit"],noise_chan=qsim_noise_opseq)
			elif op["op"] == "measure":
				qc.qmeasure(op["qubits"],cbit_list=op["clbits"])
			elif op["op"] == "noise":
				qckt_noise_op = op['noiseChan']
				qsim_noise_opseq = convert_to_qsim_noise_op_sequence(op["noiseChan"])
				qc.qnoise(qsim_noise_opseq, op['qubits'])
			elif op["op"] == "probe":
				nqbits = job.nqubits
				ncbits = job.nclbits
				(creglist, sveclist, cregcounts, runstats) = qc.qsnapshot()
				print('\n'+op["header"])
				for i in range(len(sveclist)):
					if (op["probestates"] is None and abs(sveclist[i][i]) > self.maxerr) \
							or (op["probestates"] is not None and i in op["probestates"]):
						print(("{0:0"+str(nqbits)+"b}    ... | {1:.8f}").format(i, sveclist[i][i]))
				cregsz = len(creglist)
				print("CREGISTER: ",end="")
				for i in range(cregsz):
					print("{:01b}".format(creglist[i]),end="")
				print()
			elif op["op"] == "noop":
				pass
			else:
				raise QCktException("Encountered unknown instruction: "+str(op["op"]))
		# at the end of execution perform READOUT sampling the state as many times as job.shots
		qc.qreadout(nshots=job.shots)  # this will populate the creg_counts values internally, which are returned by qsnapshot()

		# Now create the Result object from the final snapshot and store it in job.result
		cregvec,statevecarr,cregcounts,runstats = qc.qsnapshot()
		cregres = Cregister()
		cregres.setvalue_vec(cregvec)
		statevec = StateVector()
		statevec.value = statevecarr
		job.result = Result(cregvals=cregres,svecvals=statevec, cregcounts=cregcounts)
		job.runstats = runstats
		return self

class Qeng:
	def __init__(self):
		pass

	def runjob(self, job):
		# cregres_list = [None]*job.shots
		# for shot_count in range(job.shots):
		qc = qsim.QSimulator(job.nqubits,job.nclbits,qtrace=False,verbose=job.verbose)
		for op in job.assembledCkt:
			if op["op"] == "gate":
				if issubclass(type(op["qubits"][0]),list) and len(op["qubits"]) == 1:
					for q in op["qubits"][0]:
						qc.qgate([op["name"],op["opMatrix"]], [q], ifcbit=op["ifcbit"])
				else:
					qc.qgate([op["name"],op["opMatrix"]], op["qubits"], ifcbit=op["ifcbit"])
			elif op["op"] == "measure":
				qc.qmeasure(op["qubits"],cbit_list=op["clbits"])
			elif op["op"] == "noise":
				pass
			elif op["op"] == "probe":
				pass
			elif op["op"] == "noop":
				pass
			else:
				raise QCktException("Encountered unknown instruction: "+str(op["op"]))
		# at the end of execution perform READOUT sampling the state as many times as job.shots
		qc.qreadout(nshots=job.shots)  # this will populate the creg_counts values internally, which are returned by qsnapshot()

		# Now create the Result object from the final snapshot and store it in job.result
		cregvec,statevecarr,cregcounts,runstats = qc.qsnapshot()
		cregres = Cregister()
		cregres.setvalue_vec(cregvec)
		# store the last statevec as well, helps with debugging
		statevec = StateVector()
		statevec.value = statevecarr
		job.result = Result(cregvals=cregres, cregcounts=cregcounts)
		job.runstats = runstats
		return self

class Qdeb:
	def __init__(self):
		self.maxerr = 10**(-6)
		pass

	def runjob(self, job):
		qc = qsim.QSimulator(job.nqubits,job.nclbits,qtrace=job.qtrace,verbose=job.verbose)
		for op in job.assembledCkt:
			if op["op"] == "gate":
				if issubclass(type(op["qubits"][0]),list) and len(op["qubits"]) == 1:
					for q in op["qubits"][0]:
						qc.qgate([op["name"],op["opMatrix"]], [q], ifcbit=op["ifcbit"])
				else:
					qc.qgate([op["name"],op["opMatrix"]], op["qubits"], ifcbit=op["ifcbit"])
			elif op["op"] == "measure":
				qc.qmeasure(op["qubits"],cbit_list=op["clbits"])
			elif op["op"] == "noise":
				pass
			elif op["op"] == "probe":
				nqbits = job.nqubits
				ncbits = job.nclbits
				(creglist, sveclist, cregcounts, runstats) = qc.qsnapshot()
				print('\n'+op["header"])
				for i in range(len(sveclist)):
					if (op["probestates"] is None and abs(sveclist[i]) > self.maxerr) \
							or (op["probestates"] is not None and i in op["probestates"]):
						print(("{0:0"+str(nqbits)+"b}    {1:.8f}").format(i, sveclist[i]))
				cregsz = len(creglist)
				print("CREGISTER: ",end="")
				for i in range(cregsz):
					print("{:01b}".format(creglist[i]),end="")
				print()
			elif op["op"] == "noop":
				pass
			else:
				raise QCktException("Encountered unknown instruction: "+str(op["op"]))
		# at the end of execution perform READOUT sampling the state as many times as job.shots
		qc.qreadout(nshots=job.shots)  # this will populate the creg_counts values internally, which are returned by qsnapshot()

		# Now create the Result object from the final snapshot and store it in job.result
		cregvec,statevecarr,cregcounts,runstats = qc.qsnapshot()
		cregres = Cregister()
		cregres.setvalue_vec(cregvec)
		statevec = StateVector()
		statevec.value = statevecarr
		job.result = Result(cregvals=cregres, svecvals=statevec, cregcounts=cregcounts)
		job.runstats = runstats
		return self


class qsimSvc(BackendSvc):
	def __init__(self,connectionToken=None):
		# Most other services will actually open a connection to the service provider, authenticate, store the seeion object, etc.
		pass
	def listInstances(self):
		# Most other services will implement the protocol to reach the service through the session object, and fetch the list of instances
		return list(qsimSvc.instances.keys())
	def getInstance(self, instkey):
		if instkey in qsimSvc.instances.keys():
			retinst = qsimSvc.instances[instkey]
			return retinst()
		else:
			raise QCktException("No such quantum computing instance "+str(instkey))
	# in our case since the list of instances is fixed we have simply statically initialized it
	instances = {"qsim-eng":Qeng,"qsim-deb":Qdeb, 'dmqsim-eng':DMQeng, "dmqsim-deb":DMQdeb}




from setuptools import setup, find_packages

with open('RELEASE-splash.md', 'r') as file:
	long_description = file.read()

setup(
    name='qucircuit',
    version='3.1',
    license='GPL',
    author='Atul Varshneya',
    author_email='atul.varshneya@gmail.com',
	description='Quantum computer programming with circuits paradigm',
	long_description=long_description,
	long_description_content_type='text/markdown',
    packages=find_packages('.'),
    # package_dir={'': '.'},
    keywords='qsystems quantum computing',
	url='https://github.com/atulvarshneya/quantum-computing',
	python_requires='>=3.11',
    install_requires=[
        'numpy',
        'matplotlib',
        'qusimulator>=3.0'
        ]
    )



from .noiseUtils import NoiseChannel, \
    NoiseChannelSequence, \
    NoiseChannelApplierSequense, \
    NoiseProfile, \
    consolidate_gate_noise

from .noiseChannels import bit_flip, \
    phase_flip, \
    depolarizing, \
    amplitude_damping, \
    phase_damping, \
    pauli_channel, \
    generalized_amplitude_damping, \
    dummy_2qubit_chan, \
    two_qubit_dephasing, \
    two_qubit_depolarizing, \
    noise_channel_list, \
    noise_channel_lookup
